﻿using System.Net;
using BuildSoft.OscCore;

namespace VRC_Console;

public class OSCCollar
{

    public static void Main(string[] args)
    {
        OSCCollar collar;
        if (args.Length == 3)
        {
            string ip = args[0];
            if(!IPAddress.TryParse(args[0], out IPAddress? test) ||
               !int.TryParse(args[1], out var portReceive) ||
               !int.TryParse(args[2], out var portSend))
            {
                Console.WriteLine("Invalid parameter.\n\rParameters <ip> <receivePort> <sendPort>");
                return;
            }
            collar = new OSCCollar(ip, portReceive, portSend);
        }else
            collar = new OSCCollar();
    }
    
    private OscServer Server { get; init; }
    private OscClient Client { get; init; }
    private float _collarAngle, _collarStretch;
    private double _leftOrRight = 1;
    private double _verticalMovement = 0;
    private double _horizontalMovement = 0;
    private bool _allowMoving = false;
    private uint _nilSent = 0;
    private DateTime _lastNilMessageSent = DateTime.UnixEpoch;
    private DateTime _lastConsoleOutput = DateTime.UnixEpoch;
    private static readonly TimeSpan ConsoleUpdateInterval = TimeSpan.FromMilliseconds(100);
    private static readonly TimeSpan UpdateFieldsTimeout = TimeSpan.FromMilliseconds(1);
    private static readonly TimeSpan NilMessageMaxTimeout = TimeSpan.FromMilliseconds(400);

    private OSCCollar(string ip = "127.0.0.1", int portReceive = 9001, int portSend = 9000)
    {
        this.Server = new OscServer(portReceive);
        this.Client = new OscClient(ip, portSend);
        this._collarAngle = 0;
        this._collarStretch = 0;
        Server.TryAddMethod("/avatar/parameters/Collar_Angle", CollarAngleHandle);
        Server.TryAddMethod("/avatar/parameters/CollarRight", CollarLeftRightHandle);
        Server.TryAddMethod("/avatar/parameters/Collar_Allow_Moving", AllowMovingHandle);
        Server.TryAddMethodPair("/avatar/parameters/Collar_Stretch", CollarStretchHandle, CalculateMovement);
        this.Server.Start();

        Thread runningThread = new Thread(RunningThread);
        runningThread.Start();
    }

    private void PrintOutput(double verticalMovement, double horizontalMovement)
    {
        Console.Clear();
        Console.WriteLine($"OSC Collar - Status: {(_allowMoving ? "enabled" : "disabled")}");
        Console.WriteLine("===============================================================");
        Console.WriteLine($"Angle:...............{_collarAngle:0.0000}");
        Console.WriteLine($"Left/Right:..........{(_leftOrRight < 0 ? "left" : "right")}");
        Console.WriteLine($"Stretch:.............{_collarStretch:0.0000}");
        Console.WriteLine($"Vertical Movement:...{verticalMovement:0.0000}");
        Console.WriteLine($"Horizontal Movement:.{horizontalMovement:0.0000}");
        Console.SetCursorPosition(0, Console.WindowHeight - 2);
        Console.Write($"/input nil sent {_nilSent}");
        
        Console.SetCursorPosition(28, 8);
        Console.WriteLine("----------+----------");
        for (int i = 4; i < 13; i++)
        {
            if(i == 8)
                continue;
            Console.SetCursorPosition(38,i);
            Console.Write("|");
        }

        int centerX = 38;
        int centerY = 8;
        int consoleX = Convert.ToInt32(Math.Floor(_horizontalMovement * 10));
        int consoleY = Convert.ToInt32(-Math.Floor(_verticalMovement * 5));
        double position = (_verticalMovement * 100) % 10;
        char c = position < 3 ? '.' : position > 7 ? '\'' : 'x';
        Console.SetCursorPosition(centerX + consoleX, centerY + consoleY);
        Console.ForegroundColor = ConsoleColor.Cyan;
        Console.Write(c);
        Console.SetCursorPosition(Console.WindowWidth - 1, Console.WindowHeight - 1);
    }

    private void AllowMovingHandle(OscMessageValues messageValues)
    {
        this._allowMoving = messageValues.ReadBooleanElement(0);
    }

    private void CollarAngleHandle(OscMessageValues messageValues)
    {
        this._collarAngle = messageValues.ReadFloatElement(0);
    }

    private void CollarLeftRightHandle(OscMessageValues messageValues)
    {
        this._leftOrRight = messageValues.ReadBooleanElement(0) ? 1 : -1;
    }

    private void CollarStretchHandle(OscMessageValues messageValues)
    {
        this._collarStretch = messageValues.ReadFloatElement(0);
    }

    private const float PowerAdjust = 0.25f;
    private void CalculateMovement()
    {
        float degrees = _collarAngle * 180;
        double radians = degrees * Math.PI / 180;
        _verticalMovement = Math.Cos(radians) * this._collarStretch * PowerAdjust;
        _horizontalMovement = Math.Sin(radians) * this._collarStretch * PowerAdjust * _leftOrRight;
    }

    private void RunningThread()
    {
        while (true)
        {
            this.Server.Update();
            if (_lastNilMessageSent.Add(NilMessageMaxTimeout) < DateTime.Now)
            {
                this.Client.Send("/input/Vertical", 0f);
                this.Client.Send("/input/Horizontal", 0f);
                this._lastNilMessageSent = DateTime.Now;
                _nilSent++;
            }

            if (_allowMoving)
            {
                Thread.Sleep(1);
                this.Client.Send("/input/Vertical", Convert.ToSingle(_verticalMovement));
                this.Client.Send("/input/Horizontal", Convert.ToSingle(_horizontalMovement));
            }

            if (_lastConsoleOutput.Add(ConsoleUpdateInterval) < DateTime.Now)
            {
                PrintOutput(_verticalMovement, _horizontalMovement);
                _lastConsoleOutput = DateTime.Now;
            }
            
            Thread.Sleep(UpdateFieldsTimeout);
        }
    }
}